package com.tzld.piaoquan.recommend.feature.service;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.feature.util.JSONUtils;
import lombok.extern.slf4j.Slf4j;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;

import java.util.concurrent.TimeUnit;
import java.util.function.Function;

/**
 * @author dyp
 */
@Slf4j
public abstract class AbstractFeatureService<K, V> {
    @Qualifier("redisTemplate")
    @Autowired
    private RedisTemplate<String, String> redisTemplate;

    @Qualifier("tairTemplate")
    @Autowired
    private RedisTemplate<String, String> tairTemplate;

    protected String emptyData;
    // 单位秒
    protected long emptyDataExpire;
    protected long defaultExpire;
    protected Function<K, String> keyFunc;
    protected TypeToken<V> typeToken;

    private int defaultMaximumSize = 10000;
    // 单位秒
    private int defaultRefreshAfterWrite = 60;
    private int defaultExpireAfterWrite = 60;
    private int defaultExpireAfterAccess = 60;
    private LoadingCache<K, V> cache;

    public AbstractFeatureService() {
        this.emptyData = "{}";
        // 单位秒
        this.emptyDataExpire = 60;
        this.defaultExpire = 7200;
        this.keyFunc = k -> String.valueOf(k);

        initLocalCache(defaultMaximumSize, defaultRefreshAfterWrite, defaultExpireAfterWrite, defaultExpireAfterAccess);
    }

    protected void initLocalCache(int maximumSize, int refreshAfterWrite, int expireAfterWrite,
                                  int expireAfterAccess) {
        cache = CacheBuilder.newBuilder()
                .maximumSize(maximumSize)
                .refreshAfterWrite(refreshAfterWrite, TimeUnit.SECONDS)
                .expireAfterWrite(expireAfterWrite, TimeUnit.SECONDS)
                .expireAfterAccess(expireAfterAccess, TimeUnit.SECONDS)
                .build(new CacheLoader<K, V>() {
                    @Override
                    public V load(K key) {
                        return AbstractFeatureService.this.load(key);
                    }
                });
    }

    protected V get(K key) {
        assert cache != null;
        try {
            return cache.get(key);
        } catch (Exception e) {
            log.error("get local cache error", e);
        }
        return null;
    }


    private V load(K key) {
        V feature = getFromCache(key);
        if (feature == null) {
            feature = getFromSource(key);
            // TODO 可异步
            saveToCache(key, feature);
        }
        return feature;
    }

    private V getFromCache(K key) {
        String redisKey = keyFunc.apply(key);
        String value = redisTemplate.opsForValue().get(redisKey);
        return JSONUtils.fromJson(value, typeToken, null);
    }

    private V getFromSource(K key) {
        String sourceKey = cacheKey(key);
        String value = tairTemplate.opsForValue().get(sourceKey);
        return JSONUtils.fromJson(value, typeToken, null);
    }

    private void saveToCache(K key, V value) {
        String cacheKey = cacheKey(key);
        String cacheValue = value == null
                ? emptyData
                : JSONUtils.toJson(value);
        long expire = value == null
                ? emptyDataExpire
                : defaultExpire;
        // TODO 评估过期时间
        redisTemplate.opsForValue().set(cacheKey, cacheValue, expire, TimeUnit.SECONDS);
    }


    private String cacheKey(K key) {
        return keyFunc.apply(key);
    }

    protected void save(K key, V value) {
        saveToSource(key, value);
        // cache aside
        deleteFromCache(key);
    }

    private void saveToSource(K key, V value) {
        String sourceKey = cacheKey(key);
        String sourceValue = JSONUtils.toJson(value);
        tairTemplate.opsForValue().set(sourceKey, sourceValue);
    }

    private void deleteFromCache(K key) {
        String cacheKey = cacheKey(key);
        redisTemplate.delete(cacheKey);
    }

}
